# Copyright 2015, Pinterest, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Validation tests for the job executor."""
import mock
import os
import subprocess
import unittest

from pinball.config.pinball_config import PinballConfig
from pinball.workflow.event import Event
from pinball.workflow.job import ShellJob
from pinball.workflow.job_executor import ExecutionRecord
from pinball.workflow.job_executor import ShellJobExecutor


__author__ = 'Pawel Garbacki'
__copyright__ = 'Copyright 2015, Pinterest, Inc.'
__credits__ = [__author__]
__license__ = 'Apache'
__version__ = '2.0'


class ShellJobExecutorTestCase(unittest.TestCase):
    def setUp(self):
        self._data_builder = mock.Mock()
        self._emailer = mock.Mock()
        job = ShellJob(name='some_job',
                       command='printf "line1\\nline2\\nline3";'
                               'printf "line1\\nline2" >&2',
                       emails=['some_email@pinterest.com'],
                       warn_timeout_sec=10,
                       abort_timeout_sec=20)
        self._executor = ShellJobExecutor('some_workflow', '123', 'some_job',
                                          job, self._data_builder,
                                          self._emailer)
        # Set PinballConfig to enable s3 log saver
        PinballConfig.S3_LOGS_DIR_PREFIX = 's3n://pinball/tmp/'
        PinballConfig.S3_LOGS_DIR = \
            PinballConfig.S3_LOGS_DIR_PREFIX \
            + PinballConfig.JOB_LOG_PATH_PREFIX


    @mock.patch('pinball.workflow.log_saver.S3FileLogSaver._get_or_create_s3_key')
    @mock.patch('__builtin__.open')
    def test_token_lost(self, open_mock, get_s3_key_mock):
        file_mock = mock.MagicMock()
        open_mock.return_value = file_mock
        file_mock.__enter__.return_value = file_mock

        s3_key_mock = mock.MagicMock()
        get_s3_key_mock.return_value = s3_key_mock
        s3_key_mock.__enter__.return_value = s3_key_mock

        execution_record = ExecutionRecord(start_time=10)
        self._executor.job.history = [execution_record]

        self.assertFalse(self._executor.prepare())
        file_mock.write.assert_called_once_with('executor failed to renew job '
                                                'ownership on time\n')
        get_s3_key_mock.assert_called_once_with('s3n://pinball/tmp/pinball_job_logs/'
                                                'some_workflow/123/some_job.10.pinlog')

    @mock.patch('pinball.workflow.log_saver.S3FileLogSaver.open')
    @mock.patch('os.path.exists')
    @mock.patch('__builtin__.open')
    def test_events(self, open_mock, exists_mock, s3_open_mock):
        file_mock = mock.MagicMock()
        open_mock.return_value = file_mock
        file_mock.__enter__.return_value = file_mock

        exists_mock.return_value = True
        some_event = Event(creator='some_creator')
        some_other_event = Event(creator='some_other_creator')
        self._executor.job.events = [some_event, some_other_event]
        self.assertTrue(self._executor.prepare())
        self.assertEqual(1, len(self._executor.job.history))
        execution_record = self._executor.job.history[0]
        self.assertEqual([some_event, some_other_event],
                         execution_record.events)
        self.assertEqual(s3_open_mock.call_count, 2)

    def test_disabled(self):
        self._executor.job.disabled = True
        self.assertTrue(self._executor.prepare())
        self.assertEqual(1, len(self._executor.job.history))
        execution_record = self._executor.job.history[0]
        self.assertEqual('DISABLED', execution_record.info)

    @mock.patch('subprocess.Popen')
    def test_execute_cleanup(self, subprocess_mock):

        self._executor.job.cleanup_template = 'cleanup %(job_id)s'
        execution_record = ExecutionRecord()
        execution_record.properties['job_id'] = '123'
        self._executor.job.history = [execution_record]
        self._executor._execute_cleanup()
        env = os.environ.copy()
        env.pop('DJANGO_SETTINGS_MODULE', None)
        subprocess_mock.assert_called_with('cleanup 123',
                                           stdout=subprocess.PIPE,
                                           stderr=subprocess.PIPE,
                                           shell=True,
                                           env=env,
                                           preexec_fn=os.setsid)

    @mock.patch('time.time')
    def test_check_timeout_noop(self, time_mock):
        execution_record = ExecutionRecord(start_time=10)
        self._executor.job.history = [execution_record]

        time_mock.return_value = 15
        self._executor._check_timeouts()
        self.assertEqual(
            0, self._emailer.send_job_timeout_warning_message.call_count)

        time_mock.return_value = 25
        self._data_builder.get_schedule.return_value = None
        job_execution_data = mock.Mock()
        self._data_builder.get_execution.return_value = job_execution_data
        self._executor._check_timeouts()
        self._data_builder.get_schedule.assert_called_once_with(
            'some_workflow')
        self._data_builder.get_execution.assert_called_once_with(
            'some_workflow', '123', 'some_job', 0)
        self._emailer.send_job_timeout_warning_message.assert_called_once_with(
            ['some_email@pinterest.com'], job_execution_data)

        time_mock.return_value = 35
        self._executor._check_timeouts()
        self.assertTrue(self._executor._aborted)

    @mock.patch('pinball.workflow.log_saver.S3FileLogSaver._get_or_create_s3_key')
    @mock.patch('os.path.exists')
    @mock.patch('__builtin__.open')
    def test_execute(self, open_mock, exists_mock, get_s3_key_mock):
        file_mock = mock.MagicMock()
        open_mock.return_value = file_mock
        file_mock.__enter__.return_value = file_mock

        s3_key_mock = mock.MagicMock()
        get_s3_key_mock.return_value = s3_key_mock
        s3_key_mock.__enter__.return_value = s3_key_mock

        self.assertTrue(self._executor.prepare())
        self.assertTrue(self._executor.execute())

        file_mock.write.assert_has_calls(
            [mock.call('line1\n'), mock.call('line2\n'), mock.call('line3'),
             mock.call('line1\n'), mock.call('line2')],
            any_order=True)
        self.assertEqual(file_mock.write.call_count, 5)

        self.assertEqual(1, len(self._executor.job.history))
        execution_record = self._executor.job.history[0]
        self.assertEqual(0, execution_record.exit_code)

        self.assertEqual(2, get_s3_key_mock.call_count)

    @mock.patch('pinball.workflow.log_saver.S3FileLogSaver._get_or_create_s3_key')
    @mock.patch('os.makedirs')
    @mock.patch('__builtin__.open')
    def test_abort(self, open_mock, _, get_s3_key_mock):
        file_mock = mock.MagicMock()
        open_mock.return_value = file_mock
        file_mock.__enter__.return_value = file_mock

        s3_key_mock = mock.MagicMock()
        get_s3_key_mock.return_value = s3_key_mock
        s3_key_mock.__enter__.return_value = s3_key_mock

        self.assertTrue(self._executor.prepare())
        self._executor.abort()
        self.assertFalse(self._executor.execute())

        self.assertEqual(1, len(self._executor.job.history))
        execution_record = self._executor.job.history[0]
        self.assertEqual(1, execution_record.exit_code)

    @mock.patch('pinball.workflow.log_saver.S3FileLogSaver._get_or_create_s3_key')
    @mock.patch('os.path.exists')
    @mock.patch('__builtin__.open')
    def test_execute_long_line(self, open_mock, exists_mock, get_s3_key_mock):
        file_mock = mock.MagicMock()
        open_mock.return_value = file_mock
        file_mock.__enter__.return_value = file_mock

        s3_key_mock = mock.MagicMock()
        get_s3_key_mock.return_value = s3_key_mock
        s3_key_mock.__enter__.return_value = s3_key_mock

        job = ShellJob(name='some_job',
                       command="printf \"%s\"" % ('a' * 20000),
                       emails=['some_email@pinterest.com'],
                       warn_timeout_sec=10,
                       abort_timeout_sec=20)
        executor = ShellJobExecutor('some_workflow', '123', 'some_job',
                                    job, self._data_builder,
                                    self._emailer)

        self.assertTrue(executor.prepare())
        self.assertTrue(executor.execute())

        file_mock.write.assert_has_calls(
            [mock.call('a' * 16384), mock.call('a' * 3616)])

        self.assertEqual(1, len(executor.job.history))
        execution_record = executor.job.history[0]
        self.assertEqual(0, execution_record.exit_code)

        self.assertEqual(2, get_s3_key_mock.call_count)



# Copyright 2015, Pinterest, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Validation tests for the signaller."""
import mock
import pickle
import unittest

from pinball.config.pinball_config import PinballConfig
from pinball.master.factory import Factory
from pinball.master.thrift_lib.ttypes import ModifyRequest
from pinball.master.thrift_lib.ttypes import Token
from pinball.workflow.name import Name
from pinball.workflow.signaller import Signal
from pinball.workflow.signaller import Signaller
from tests.pinball.persistence.ephemeral_store import EphemeralStore


__author__ = 'Pawel Garbacki'
__copyright__ = 'Copyright 2015, Pinterest, Inc.'
__credits__ = [__author__]
__license__ = 'Apache'
__version__ = '2.0'


class SignallerTestCase(unittest.TestCase):
    def setUp(self):
        self._factory = Factory()
        self._factory.create_master(EphemeralStore())

    def _post_signal_tokens(self):
        """Add some signal tokens to the master."""
        request = ModifyRequest(updates=[])

        signal = Signal(action=Signal.EXIT)
        name = Name(signal='exit')
        signal_token = Token(name=name.get_signal_token_name())
        signal_token.data = pickle.dumps(signal)
        request.updates.append(signal_token)

        signal = Signal(action=Signal.DRAIN)
        name.signal = 'drain'
        name.workflow = 'some_workflow'
        signal_token = Token(name=name.get_signal_token_name())
        signal_token.data = pickle.dumps(signal)
        request.updates.append(signal_token)

        name.instance = '123'
        signal_token = Token(name=name.get_signal_token_name())
        signal_token.data = pickle.dumps(signal)
        request.updates.append(signal_token)

        signal = Signal(action=Signal.ABORT)
        name.signal = 'abort'
        signal_token = Token(name=name.get_signal_token_name())
        signal_token.data = pickle.dumps(signal)
        request.updates.append(signal_token)

        client = self._factory.get_client()
        client.modify(request)

    def test_is_action_set(self):
        client = self._factory.get_client()
        signaller = Signaller(client)
        self.assertFalse(signaller.is_action_set(Signal.EXIT))
        self.assertFalse(signaller.is_action_set(Signal.DRAIN))
        self.assertFalse(signaller.is_action_set(Signal.ABORT))

        self._post_signal_tokens()

        signaller = Signaller(client)
        self.assertTrue(signaller.is_action_set(Signal.EXIT))
        self.assertFalse(signaller.is_action_set(Signal.DRAIN))
        self.assertFalse(signaller.is_action_set(Signal.ABORT))

        signaller = Signaller(client, workflow='some_workflow')
        self.assertTrue(signaller.is_action_set(Signal.EXIT))
        self.assertTrue(signaller.is_action_set(Signal.DRAIN))
        self.assertFalse(signaller.is_action_set(Signal.ABORT))

        signaller = Signaller(client, workflow='some_workflow', instance='123')
        self.assertTrue(signaller.is_action_set(Signal.EXIT))
        self.assertTrue(signaller.is_action_set(Signal.DRAIN))
        self.assertTrue(signaller.is_action_set(Signal.ABORT))

    def test_set_action(self):
        client = self._factory.get_client()

        writing_signaller = Signaller(client)
        writing_signaller.set_action(Signal.EXIT)
        reading_signaller = Signaller(client)
        # New generation.
        self.assertFalse(reading_signaller.is_action_set(Signal.EXIT))
        # Old generation.
        with mock.patch('pinball.workflow.signaller.PinballConfig.GENERATION', 0):
            self.assertTrue(reading_signaller.is_action_set(Signal.EXIT))
        self.assertFalse(reading_signaller.is_action_set(Signal.DRAIN))
        self.assertFalse(reading_signaller.is_action_set(Signal.ABORT))

        writing_signaller = Signaller(client, workflow='some_workflow')
        writing_signaller.set_action(Signal.DRAIN)
        reading_signaller = Signaller(client, workflow='some_workflow')
        # Old generation.
        with mock.patch('pinball.workflow.signaller.PinballConfig.GENERATION', 0):
            self.assertTrue(reading_signaller.is_action_set(Signal.EXIT))
        self.assertTrue(reading_signaller.is_action_set(Signal.DRAIN))
        self.assertFalse(reading_signaller.is_action_set(Signal.ABORT))

        writing_signaller = Signaller(client, workflow='some_workflow',
                                      instance='123')
        writing_signaller.set_action(Signal.ABORT)
        reading_signaller = Signaller(client, workflow='some_workflow',
                                      instance='123')
        # Old generation.
        with mock.patch('pinball.workflow.signaller.PinballConfig.GENERATION', 0):
            self.assertTrue(reading_signaller.is_action_set(Signal.EXIT))
        self.assertTrue(reading_signaller.is_action_set(Signal.DRAIN))
        self.assertTrue(reading_signaller.is_action_set(Signal.ABORT))

    def test_remove_sction(self):
        client = self._factory.get_client()

        writing_signaller = Signaller(client)
        writing_signaller.set_action(Signal.EXIT)
        writing_signaller.remove_action(Signal.EXIT)
        self.assertFalse(writing_signaller.is_action_set(Signal.EXIT))
        reading_signaller = Signaller(client)
        self.assertFalse(reading_signaller.is_action_set(Signal.EXIT))

        writing_signaller = Signaller(client, workflow='some_workflow')
        writing_signaller.set_action(Signal.DRAIN)
        reading_signaller = Signaller(client, workflow='some_workflow')
        self.assertTrue(reading_signaller.is_action_set(Signal.DRAIN))
        writing_signaller.remove_action(Signal.DRAIN)
        self.assertFalse(writing_signaller.is_action_set(Signal.DRAIN))
        reading_signaller = Signaller(client)
        self.assertFalse(reading_signaller.is_action_set(Signal.DRAIN))

        writing_signaller = Signaller(client, workflow='some_workflow',
                                      instance='123')
        writing_signaller.set_action(Signal.ABORT)
        reading_signaller = Signaller(client, workflow='some_workflow',
                                      instance='123')
        self.assertTrue(reading_signaller.is_action_set(Signal.ABORT))
        writing_signaller.remove_action(Signal.ABORT)
        self.assertFalse(writing_signaller.is_action_set(Signal.ABORT))
        reading_signaller = Signaller(client)
        self.assertFalse(reading_signaller.is_action_set(Signal.ABORT))

    def test_get_attribute(self):
        client = self._factory.get_client()

        writing_signaller = Signaller(client)
        writing_signaller.set_action(Signal.EXIT)
        self.assertEqual(PinballConfig.GENERATION,
                         writing_signaller.get_attribute(
                             Signal.EXIT,
                             Signal.GENERATION_ATTR))
        reading_signaller = Signaller(client)
        self.assertEqual(PinballConfig.GENERATION,
                         reading_signaller.get_attribute(
                             Signal.EXIT,
                             Signal.GENERATION_ATTR))

    def test_set_attribute_if_missing(self):
        client = self._factory.get_client()

        writing_signaller = Signaller(client, workflow='some_workflow',
                                      instance='123')

        self.assertFalse(writing_signaller.set_attribute_if_missing(
                         Signal.ARCHIVE, Signal.TIMESTAMP_ATTR, 12345))

        writing_signaller.set_action(Signal.ARCHIVE)
        self.assertTrue(writing_signaller.set_attribute_if_missing(
                        Signal.ARCHIVE, Signal.TIMESTAMP_ATTR, 12345))
        self.assertEqual(12345,
                         writing_signaller.get_attribute(
                             Signal.ARCHIVE,
                             Signal.TIMESTAMP_ATTR))

        self.assertFalse(writing_signaller.set_attribute_if_missing(
                         Signal.ARCHIVE, Signal.TIMESTAMP_ATTR, 123456))

        reading_signaller = Signaller(client, workflow='some_workflow',
                                      instance='123')
        self.assertEqual(12345,
                         reading_signaller.get_attribute(
                             Signal.ARCHIVE,
                             Signal.TIMESTAMP_ATTR))



# Copyright 2015, Pinterest, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Generic ui-related utilities."""
import pickle

from pinball.config.pinball_config import PinballConfig
from pinball.ui.data import JobData
from pinball.ui.data import Status
from pinball.parser.config_parser import ParserCaller
from pinball.parser.utils import load_parser_with_caller
from pinball.workflow.name import Name


__author__ = 'Pawel Garbacki, Mao Ye'
__copyright__ = 'Copyright 2015, Pinterest, Inc.'
__credits__ = [__author__]
__license__ = 'Apache'
__version__ = '2.0'


def get_workflow_jobs_from_parser(workflow):
    config_parser = load_parser_with_caller(PinballConfig.PARSER,
                                            PinballConfig.PARSER_PARAMS,
                                            ParserCaller.UI)
    tokens = config_parser.get_workflow_tokens(workflow)
    jobs_data = []
    for token in tokens:
        name = Name.from_job_token_name(token.name)
        if name.job:
            assert name.workflow == workflow
            job = pickle.loads(token.data)
            jobs_data.append(JobData(workflow=workflow,
                                     instance=None,
                                     job=name.job,
                                     job_type=job.__class__.__name__,
                                     is_condition=job.IS_CONDITION,
                                     info=job.info(),
                                     inputs=job.inputs,
                                     outputs=job.outputs,
                                     emails=job.emails,
                                     max_attempts=job.max_attempts,
                                     retry_delay_sec=job.retry_delay_sec,
                                     warn_timeout_sec=job.warn_timeout_sec,
                                     abort_timeout_sec=job.abort_timeout_sec,
                                     priority=token.priority,
                                     status=Status.NEVER_RUN))
    return jobs_data



# Copyright 2015, Pinterest, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Inspector supports traversal of the token tree hierarchy.

In particular, inspector comes handy when we work with tokens hierarchies that
contain name components unknown to the client.  Inspector understands the
hierarchical names of workflow tokens and exposes an interface to traverse
those names level-by-level.  E.g., we may use the inspector to find all
workflow instance ids or all waiting jobs in a given workflow isntance.
"""


from pinball.master.thrift_lib.ttypes import GroupRequest
from pinball.workflow.name import Name


__author__ = 'Pawel Garbacki'
__copyright__ = 'Copyright 2015, Pinterest, Inc.'
__credits__ = [__author__]
__license__ = 'Apache'
__version__ = '2.0'


class Inspector(object):
    def __init__(self, client):
        self._client = client

    def get_workflow_names(self):
        """Return list of workflow names."""
        request = GroupRequest()
        request.namePrefix = Name.WORKFLOW_PREFIX
        request.groupSuffix = Name.DELIMITER
        response = self._client.group(request)
        workflow_names = []
        if response.counts:
            for prefix in response.counts.keys():
                name = Name.from_workflow_prefix(prefix)
                if name.workflow:
                    workflow_names.append(name.workflow)
        return workflow_names

    def get_workflow_instances(self, workflow_name):
        """Return list of instances of a given workflow."""
        request = GroupRequest()
        name = Name()
        name.workflow = workflow_name
        request.namePrefix = name.get_workflow_prefix()
        request.groupSuffix = Name.DELIMITER
        response = self._client.group(request)
        instance_names = []
        if response.counts:
            for prefix in response.counts.keys():
                name = Name.from_instance_prefix(prefix)
                if name.instance:
                    instance_names.append(name.instance)
        return instance_names

    def _get_job_names(self, workflow_name, instance, state):
        """Return list of job names in a given workflow instance and state.

        E.g., assume the following tokens are stored in the master:
            /workflow/some_workflow/12345/waiting/some_waiting_job
            /workflow/some_workflow/12345/waiting/some_other_waiting_job
            /workflow/some_workflow/12345/runnable/some_runnable_job

        the method called with workflow_name=some_workflow, instance=12345,
        state=waiting will return [some_waiting_job, some_other_waiting_job].
        """
        request = GroupRequest()
        name = Name()
        name.workflow = workflow_name
        name.instance = instance
        name.job_state = state
        request.namePrefix = name.get_job_state_prefix()
        request.groupSuffix = Name.DELIMITER
        response = self._client.group(request)
        job_names = []
        if response.counts:
            for job_name in response.counts.keys():
                name = Name.from_job_token_name(job_name)
                job_names.append(name.job)
        return job_names

    def get_runnable_job_names(self, workflow_name, instance):
        """Return names of runnable jobs in a given workflow instance."""
        return self._get_job_names(workflow_name,
                                   instance,
                                   Name.RUNNABLE_STATE)

    def get_waiting_job_names(self, workflow_name, instance):
        """Return names of waiting jobs in a given workflow instance."""
        return self._get_job_names(workflow_name,
                                   instance,
                                   Name.WAITING_STATE)

    def get_event_names(self, workflow_name, instance, job, input_name):
        """Return names of events under a workflow instance, job, and input."""
        request = GroupRequest()
        name = Name()
        name.workflow = workflow_name
        name.instance = instance
        name.job = job
        name.input = input_name
        request.namePrefix = name.get_input_prefix()
        request.groupSuffix = Name.DELIMITER
        response = self._client.group(request)
        events = []
        if response.counts:
            for event in response.counts.keys():
                name = Name.from_event_token_name(event)
                events.append(name.event)
        return events



# Copyright 2015, Pinterest, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

from pinball_ext.common import utils
from pinball_ext.job.basic_jobs import ClusterJob


__author__ = 'Changshu Liu, Mao Ye, Mohammad Shahangian'
__copyright__ = 'Copyright 2015, Pinterest, Inc.'
__credits__ = [__author__]
__license__ = 'Apache'
__version__ = '2.0'


LOG = utils.get_logger('pinball_ext.job.hive_jobs')


class HiveJobBase(ClusterJob):
    """Base class for jobs that run Hive query."""

    # If set to true, upload archive; otherwise don't upload archive
    _UPLOAD_ARCHIVE = False

    def _get_query_template(self):
        """Get the hive query template as a string.

        The returned template may contain some place holder parameters that will
        be replaced with self.params.
        """
        raise NotImplementedError("No query template available in HiveJobBase")

    def _setup(self):
        super(HiveJobBase, self)._setup()
        self._delay()

    def _execute(self):
        super(HiveJobBase, self)._execute()
        self._job_output, self._job_stderr, self._job_ids =\
            self.executor.run_hive_query(
                self._get_query_template() % self.params,
                upload_archive=self._UPLOAD_ARCHIVE)


class HiveJob(HiveJobBase):
    """A job to run Hive query whose template defined as job attribute.

    It retrieves job attribute: _QUERY_TEMPLATE as query template.
    """

    _QUERY_TEMPLATE = None

    def _get_query_template(self):
        if not self._QUERY_TEMPLATE:
            raise Exception('_QUERY_TEMPLATE is empty')

        return self._QUERY_TEMPLATE


class HiveFileJob(HiveJobBase):
    """A job to run Hive query whose template is stored in a file.

    The file path is specified in job attribute: _QUERY_TEMPLATE_FILE. The path
    is relative to job's root_dir attributed which is the dir where HiveFileJob
    class is defined by default. Derived job class can override _setup() method
    to override this attribute.
    """

    _QUERY_TEMPLATE_FILE = None

    def _setup(self):
        super(HiveFileJob, self)._setup()
        self.root_dir = os.path.dirname(__file__)

    def _get_query_template(self):
        if not self._QUERY_TEMPLATE_FILE:
            raise NotImplementedError('_QUERY_FILE is empty')

        query_file_path = os.path.join(self.root_dir, self._QUERY_TEMPLATE_FILE)
        LOG.info('reading hive query template from: %s ...', query_file_path)
        with open(query_file_path, 'r') as f:
            query_template = f.read()
            return query_template



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from pinball import __version__
from setuptools import setup


setup(
    name='pinball',
    version=__version__,
    description='Workflow manager and scheduler',
    author='Pawel Garbacki',
    author_email='pawel@pinterest.com',
    url='https://github.com/pinterest/pinball',
    packages=[
        'pinball',
        'pinball.authentication',
        'pinball.common',
        'pinball.config',
        'pinball.master',
        'pinball.master.thrift_lib',
        'pinball.parser',
        'pinball.persistence',
        'pinball.scheduler',
        'pinball.tools',
        'pinball.ui',
        'pinball.workflow',
        'pinball_ext',
        'pinball_ext.common',
        'pinball_ext.examples',
        'pinball_ext.executor',
        'pinball_ext.job',
        'pinball_ext.workflow',
        'tests',
        'tests.pinball',
        'tests.pinball.authentication',
        'tests.pinball.config',
        'tests.pinball.master',
        'tests.pinball.parser',
        'tests.pinball.persistence',
        'tests.pinball.scheduler',
        'tests.pinball.tools',
        'tests.pinball.ui',
        'tests.pinball.workflow',
        'tests.pinball_ext',
        'tests.pinball_ext.executor',
        'tests.pinball_ext.job',
        'tests.pinball_ext.workflow',
    ],
    package_dir={'pinball': 'pinball'},
    include_package_data=True,
    install_requires=[
        'argparse==1.2.1',
        'boto==2.8.0',
        'Django==1.5.4',
        'guppy==0.1.10',
        'mock==0.8.0',
        'MySQL-python==1.2.3',
        'oauth2client==1.2',
        'pycrypto==2.6',
        'pydot==1.0.28',
        'python-dateutil==2.4',
        'PyTrie==0.1',
        'pytz==2013b',
        'pyyaml==3.11',
        'qds_sdk==1.3.0',
        'thrift==0.8.0',
        'tox>=1.6.1',
    ],
    license="Apache 2.0",
    zip_safe=False,
    keywords='pinball, workflow manager, batch, job, hadoop, emr, qubole',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'Programming Language :: Python :: 2.7',
        'License :: OSI Approved :: Apache Software License',
        'Topic :: System :: Distributed Computing',
    ],
    test_suite='tests',
)



# Copyright 2015, Pinterest, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Validation tests for schedules."""
import pickle
import time
import unittest

from pinball.master.factory import Factory
from pinball.master.thrift_lib.ttypes import ModifyRequest
from pinball.master.thrift_lib.ttypes import Token
from pinball.scheduler.overrun_policy import OverrunPolicy
from pinball.scheduler.schedule import WorkflowSchedule
from pinball.scheduler.scheduler import Scheduler
from pinball.workflow.emailer import Emailer
from pinball.workflow.name import Name
from tests.pinball.persistence.ephemeral_store import EphemeralStore


__author__ = 'Pawel Garbacki'
__copyright__ = 'Copyright 2015, Pinterest, Inc.'
__credits__ = [__author__]
__license__ = 'Apache'
__version__ = '2.0'


# mock.Mock cannot be pickled so we have to define our own mock class.
class MockWorkflowSchedule(WorkflowSchedule):
    def __init__(self, is_running, is_failed):
        super(MockWorkflowSchedule, self).__init__(
            next_run_time=int(time.time() - 10),
            recurrence_seconds=10,
            workflow='some_workflow')
        self._is_running = is_running
        self._is_failed = is_failed
        self.abort_called = False

    def is_running(self, store):
        return self._is_running

    def is_failed(self, store):
        return self._is_failed

    def abort_running(self, client, store):
        self.abort_called = True
        return True

    def run(self, emailer, store):
        return ModifyRequest(updates=[Token()])


class SchedulerTestCase(unittest.TestCase):
    def setUp(self):
        self._factory = Factory()
        store = EphemeralStore()
        self._factory.create_master(store)
        emailer = Emailer('some_host', '8080')
        self._scheduler = Scheduler(self._factory.get_client(), store, emailer)
        self._client = self._factory.get_client()
        self._post_schedule_token()

    @staticmethod
    def _get_schedule_token():
        name = Name(workflow='workflow_0')
        now = int(time.time())
        token = Token(name=name.get_workflow_schedule_token_name(),
                      owner='some_owner',
                      expirationTime=now - 10)
        schedule = WorkflowSchedule(next_run_time=now - 10,
                                    recurrence_seconds=10,
                                    workflow='workflow_0')
        token.data = pickle.dumps(schedule)
        return token

    def _post_schedule_token(self):
        """Add schedule token to the master."""
        request = ModifyRequest()
        request.updates = [SchedulerTestCase._get_schedule_token()]
        self._client.modify(request)

    def test_own_schedule_token(self):
        self._scheduler._own_schedule_token()
        self.assertIsNotNone(self._scheduler._owned_schedule_token)

    def test_advance_schedule(self):
        self._scheduler._own_schedule_token()
        token = self._scheduler._owned_schedule_token
        owned_schedule = pickle.loads(token.data)
        self._scheduler._advance_schedule(owned_schedule)
        now = int(time.time())
        self.assertGreater(token.expirationTime, now - 10)
        schedule = pickle.loads(token.data)
        self.assertEqual(token.expirationTime, schedule.next_run_time)

    def test_run_or_reschedule_incorrect_expiration_time(self):
        self._scheduler._own_schedule_token()
        token = self._scheduler._owned_schedule_token
        schedule = pickle.loads(token.data)
        schedule.next_run_time = int(time.time() + 1000)
        token.data = pickle.dumps(schedule)
        self.assertRaises(AssertionError, self._scheduler._run_or_reschedule)

    def _run_or_reschedule(self, overrun_policy, is_running=True,
                           is_failed=True, is_abort_called=False):
        self._scheduler._own_schedule_token()
        token = self._scheduler._owned_schedule_token
        schedule = MockWorkflowSchedule(is_running, is_failed)
        schedule.overrun_policy = overrun_policy
        token.data = pickle.dumps(schedule)
        token.expirationTime = schedule.next_run_time
        old_expiration_time = token.expirationTime

        self._scheduler._run_or_reschedule()

        token = self._scheduler._owned_schedule_token
        new_expiration_time = token.expirationTime
        self.assertGreater(new_expiration_time, old_expiration_time)
        schedule = pickle.loads(token.data)
        self.assertEqual(is_abort_called, schedule.abort_called)

    def test_run_START_NEW(self):
        self._run_or_reschedule(OverrunPolicy.START_NEW)
        self.assertIsNotNone(self._scheduler._request)

    def test_reschedule_SKIP(self):
        self._run_or_reschedule(OverrunPolicy.SKIP)
        self.assertIsNone(self._scheduler._request)

    def test_run_ABORT_RUNNING(self):
        self._run_or_reschedule(OverrunPolicy.ABORT_RUNNING,
                                is_abort_called=True)
        self.assertIsNotNone(self._scheduler._request)

    def test_reschedule_DELAY(self):
        self._run_or_reschedule(OverrunPolicy.DELAY)
        token = self._scheduler._owned_schedule_token
        schedule = pickle.loads(token.data)
        self.assertLess(schedule.next_run_time, token.expirationTime)
        self.assertIsNone(self._scheduler._request)

    def test_run_DELAY(self):
        self._run_or_reschedule(OverrunPolicy.DELAY, is_running=False)
        self.assertIsNotNone(self._scheduler._request)

    def test_reschedule_DELAY_UNTIL_SUCCESS(self):
        self._run_or_reschedule(OverrunPolicy.DELAY_UNTIL_SUCCESS,
                                is_running=False, is_failed=True)
        token = self._scheduler._owned_schedule_token
        schedule = pickle.loads(token.data)
        self.assertLess(schedule.next_run_time, token.expirationTime)
        self.assertIsNone(self._scheduler._request)

    def test_run_DELAY_UNTIL_SUCCESS(self):
        self._run_or_reschedule(OverrunPolicy.DELAY_UNTIL_SUCCESS,
                                is_running=False, is_failed=False)
        self.assertIsNotNone(self._scheduler._request)



# Copyright 2015, Pinterest, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Validation tests for the job."""
import unittest

from pinball.workflow.event import Event
from pinball.workflow.job import ShellJob
from pinball.workflow.job_executor import ExecutionRecord


__author__ = 'Pawel Garbacki'
__copyright__ = 'Copyright 2015, Pinterest, Inc.'
__credits__ = [__author__]
__license__ = 'Apache'
__version__ = '2.0'


class JobTestCase(unittest.TestCase):
    def test_retry(self):
        job = ShellJob(name='some_job')

        # Empty history.
        self.assertFalse(job.retry())

        # History with a successful execution.
        record = ExecutionRecord(instance=123, exit_code=0)
        job.history.append(record)
        self.assertRaises(AssertionError, job.retry)

        # History with too many failures.
        record = ExecutionRecord(instance=1234, exit_code=1)
        job.history.append(record)
        self.assertFalse(job.retry())

        # History without too many failures.
        job.max_attempts = 2
        self.assertTrue(job.retry())

        # History with too many failures in a different instance.
        job.history.append(record)
        record = ExecutionRecord(instance=12345, exit_code=1)
        job.history.append(record)
        self.assertTrue(job.retry())


class ShellJobTestCase(unittest.TestCase):
    def test_customize_command(self):
        job = ShellJob(name='some_job',
                       inputs=['some_input', 'some_other_input'])
        some_event = Event(attributes={'some_attr': 'some_value'})
        some_other_event = Event(attributes={
            'some_attr': 'some_other_value',
            'yet_another_attr': 'yet_another_value'})
        execution_record = ExecutionRecord(instance=123, start_time=10)
        execution_record.events = [some_event, some_other_event]
        job.history = [execution_record]

        # Empty command.
        job.command = ''
        self.assertEqual('', job.customize_command())

        # Command with no attributes.
        job.command = 'some_command'
        self.assertEqual('some_command', job.customize_command())

        # Command with attributes.
        job.command = ('%(non_existent_attr)s %(some_attr)s '
                       '%(yet_another_attr)s')
        self.assertEqual(' some_value,some_other_value yet_another_value',
                         job.customize_command())

        # Command with percentage marks.
        job.command = ('%% some_command')
        self.assertEqual('% some_command', job.customize_command())



# Copyright 2015, Pinterest, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Factory for creating token master and client objects."""
import socket

from thrift.protocol import TBinaryProtocol
from thrift.server import TServer
from thrift.transport import TSocket
from thrift.transport import TTransport

from pinball.config.utils import get_log
from pinball.master.client import LocalClient, RemoteClient
from pinball.master.master_handler import MasterHandler
from pinball.master.thrift_lib.TokenMasterService import Processor


__author__ = 'Pawel Garbacki'
__copyright__ = 'Copyright 2015, Pinterest, Inc.'
__credits__ = [__author__]
__license__ = 'Apache'
__version__ = '2.0'


LOG = get_log('pinball.master.factory')


class Factory(object):
    """Factory creating token master and clients."""

    def __init__(self, master_hostname=None, master_port=None):
        """Create a factory.

        Args:
            master_hostname: Hostname of the master server.  Not required if
                master is running locally.  Defaults to the name of the local
                host.
            master_port: Port of the master server.  Not required if master
                is running locally.
        """
        self._master_handler = None
        if master_hostname:
            self._hostname = master_hostname
        else:
            self._hostname = socket.gethostname()
        self._port = master_port

    def create_master(self, store):
        """Create a local master.

        Args:
            store: The store where the master persists tokens.
        """
        self._master_handler = MasterHandler(store)

    def run_master_server(self):
        """Start thrift token master server and block waiting until it's done.
        """
        assert self._master_handler
        processor = Processor(self._master_handler)
        if self._port:
            transport = TSocket.TServerSocket(port=self._port)
        else:
            transport = TSocket.TServerSocket()
            self._port = transport.port
        tfactory = TTransport.TBufferedTransportFactory()
        pfactory = TBinaryProtocol.TBinaryProtocolFactory()
        server = TServer.TThreadedServer(processor, transport, tfactory,
                                         pfactory)

        LOG.info('Starting server on host:port %s:%d', self._hostname,
                 self._port)
        server.serve()
        LOG.info('server is done')

    def get_client(self):
        """Create local or remote client depending on the master availability.

        Return:
            A local client if this factory was used to create a master.
            Otherwise, return a thrift client connected to a remote master.  In
            the latter case, hostname and port must have been provided in the
            factory constructor.
        """
        if self._master_handler:
            return LocalClient(self._master_handler)
        else:
            if not self._hostname or not self._port:
                raise Exception('hostname and port must be defined')
            return RemoteClient(self._hostname, self._port)



# Copyright 2015, Pinterest, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Defines configurations for a workflow, class and schedule.
"""


__author__ = 'Mao Ye, Jooseong Kim'
__copyright__ = 'Copyright 2015, Pinterest, Inc.'
__credits__ = [__author__]
__license__ = 'Apache'
__version__ = '2.0'


class WorkflowConfig(object):
    """Configures a workflow.

    This class configures a workflow. This configuration includes jobs that
    belong to the workflow, the schedule the workflow runs on and the email
    addresses any notification needs to be sent to.


    Attributes:
        jobs: Dict of job name to JobConfig objects.
            Each (job name, job config) in the dict expresses one job that
            belongs to this workflow.
        final_job_config: A config for a job that will be executed as the final
            stage of the workflow. This job is executed once all other jobs of
            the workflow runs to completion successfully. This is configured by
            an instance of JobConfig.
        schedule: The schedule on which this workflow will run. This is an
            instance of ScheduleConfig.
        notify_emails: A comma separated string of email addresses. Any
            notification will be sent to these email addressses.

    """
    def __init__(self, jobs, final_job_config, schedule, notify_emails=None):
        self.jobs = jobs
        self.final_job_config = final_job_config
        self.schedule = schedule
        self.notify_emails = notify_emails


class JobConfig(object):
    """Configures a job.

    This class configures a job. It includes the job template and the name of
    other jobs that needs to successfully complete before this job starts.

    Attributes:
        template: The job template for this job. A job template defines what
            the job does.
        dependency: List of names of jobs that needs to precede this job. Only
            jobs from the same workflow can be depended on.
    """
    def __init__(self, template, dependency=None):
        self.template = template
        self.dependency = dependency


class ScheduleConfig(object):
    """Configures a schedule.

    This class configures the schedule of running this workflow.

    Attributes:
        recurrence: Recurrence or interval of running a workflow. This is
            configured by an instance of datetime.timedelta.
        reference_timestamp: The timestamp that becomes the basis of scheduling
            future jobs. Future jobs will be run periodically as specified by
            recurrence since this reference point.
        overrun_policy: Policy governing what to do if the previous workflow
            is still running. Refer to class OverrunPolicy for possible values.
        max_running_instances: Maximum number of workflow instances that can
            run at the same time.
    """
    def __init__(self, recurrence, reference_timestamp, overrun_policy=None,
                 max_running_instances=None):
        self.recurrence = recurrence
        self.reference_timestamp = reference_timestamp
        self.overrun_policy = overrun_policy
        self.max_running_instances = max_running_instances



# Copyright 2015, Pinterest, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Validation tests for the token hierarchy inspector."""
import unittest

from pinball.master.factory import Factory
from pinball.workflow.inspector import Inspector
from pinball.workflow.name import Name
from pinball.master.thrift_lib.ttypes import ModifyRequest
from pinball.master.thrift_lib.ttypes import Token
from tests.pinball.persistence.ephemeral_store import EphemeralStore


__author__ = 'Pawel Garbacki'
__copyright__ = 'Copyright 2015, Pinterest, Inc.'
__credits__ = [__author__]
__license__ = 'Apache'
__version__ = '2.0'


class InspectorTestCase(unittest.TestCase):
    def setUp(self):
        self._factory = Factory()
        self._factory.create_master(EphemeralStore())
        self._inspector = Inspector(self._factory.get_client())

    def _post_job_tokens(self):
        """Add some job tokens to the master."""
        request = ModifyRequest(updates=[])
        name = Name(workflow='some_workflow', instance='12345')
        for job_id in range(0, 2):
            if job_id % 2 == 0:
                name.job_state = Name.WAITING_STATE
            else:
                name.job_state = Name.RUNNABLE_STATE
            name.job = 'some_job_%d' % job_id
            job_token = Token(name=name.get_job_token_name())
            request.updates.append(job_token)
        client = self._factory.get_client()
        client.modify(request)

    def _post_event_tokens(self):
        """Add some event tokens to the master."""
        request = ModifyRequest(updates=[])
        name = Name(workflow='some_workflow', instance='12345')
        for job_id in range(0, 2):
            for input_id in range(0, 2):
                for event_id in range(0, 2):
                    name.job = 'some_job_%d' % job_id
                    name.input = 'some_input_%d' % input_id
                    name.event = 'some_event_%d' % event_id
                    event_token = Token(name=name.get_event_token_name())
                    request.updates.append(event_token)
        client = self._factory.get_client()
        client.modify(request)

    def test_inspect_empty_tree(self):
        self.assertEqual([], self._inspector.get_workflow_names())
        self.assertEqual([], self._inspector.get_workflow_instances(
            'some_workflow'))
        self.assertEqual([], self._inspector.get_waiting_job_names(
            'some_workflow', '12345'))
        self.assertEqual([], self._inspector.get_runnable_job_names(
            'some_workflow', '12345'))
        self.assertEqual([], self._inspector.get_event_names('some_workflow',
                                                             '12345',
                                                             'some_job_0',
                                                             'some_input_0'))

    def test_inspect_non_empty_tree(self):
        self._post_job_tokens()
        self._post_event_tokens()
        self.assertEqual(['some_workflow'],
                         self._inspector.get_workflow_names())
        self.assertEqual(['12345'], self._inspector.get_workflow_instances(
            'some_workflow'))
        self.assertEqual(['some_job_0'], self._inspector.get_waiting_job_names(
            'some_workflow', '12345'))
        self.assertEqual(['some_job_1'],
                         self._inspector.get_runnable_job_names(
                             'some_workflow', '12345'))
        event_names = self._inspector.get_event_names('some_workflow',
                                                      '12345',
                                                      'some_job_0',
                                                      'some_input_0')
        event_names = sorted(event_names)
        self.assertEqual(['some_event_0', 'some_event_1'], event_names)



# Copyright 2015, Pinterest, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utilities for interacting with Amazon s3 service."""

import re

import boto

from pinball.config.pinball_config import PinballConfig
from pinball.config.utils import get_log


__author__ = 'Mao Ye, Changshu Liu'
__copyright__ = 'Copyright 2015, Pinterest, Inc.'
__credits__ = ['Mao Ye', 'Changshu Liu']
__license__ = 'Apache'
__version__ = '2.0'


if not boto.config.has_section('Boto'):
    boto.config.add_section('Boto')
    boto.config.set('Boto', 'http_socket_timeout', '180')

LOG = get_log('pinball.common.s3_utils')


def parse_s3_location(s3_location):
    """Parse s3_location to get the bucket name and the rest of the file path.

    Args:
        s3_location: A string in the form of:
            's3n://<bucket_name>/<rest_of_the_file_path>'.

    Returns:
        bucket_name, rest_of_the_file_path
    """
    try:
        regex = r'\s*s3n://(.+?)/(.+)'
        return re.match(regex, s3_location).groups()
    except:
        raise Exception('Invalid s3 location: %s' % s3_location)


def get_s3_bucket(bucket_name):
    """Get the Boto s3 bucket reference for the given bucket_name.

    Args:
        bucket_name: name of s3 bucket.

    Returns
        s3 bucket object.
    """
    assert PinballConfig.AWS_ACCESS_KEY_ID
    assert PinballConfig.AWS_SECRET_ACCESS_KEY
    connection = boto.connect_s3(PinballConfig.AWS_ACCESS_KEY_ID,
                                 PinballConfig.AWS_SECRET_ACCESS_KEY)
    assert connection
    bucket = connection.get_bucket(bucket_name, validate=False)
    return bucket


def delete_s3_directory(s3_directory):
    """Delete the given s3 directory."""
    bucket_name, rest_of_dir_path = parse_s3_location(s3_directory)
    bucket = get_s3_bucket(bucket_name)
    rest_of_dir_path = rest_of_dir_path \
        if rest_of_dir_path[-1] == '/' else rest_of_dir_path + '/'
    bucket.delete_keys(bucket.list(prefix=rest_of_dir_path))



# Copyright 2015, Pinterest, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Implementation of workflow worker handling job execution.

Workflow is represented by a collection of job tokens.  Jobs have one or more
inputs and zero or more outputs.  An output of an upstream job connects to the
input of a downstream job.  A special type of input representing workflow
start is defined for top level jobs.  In practice, an input is a name prefix
shared by related event tokens of relevance to a specific job.

A job at a given point time is in one of two states: waiting or runnable.
All jobs start in the waiting state.  If there is at least one event in each
input of a job, the job can be made runnable.  When a worker makes a job
runnable, it consumes (i.e., removes) one event token from each of its inputs.
We call those events triggering events.

Runnable job tokens are claimed by idle workers and executed.  During execution
the ownership of the claimed job token is renewed periodically.  If a worker
fails to renew the ownership, the job token is claimed and executed by someone
else.  In this model we assume job idempotence.
"""
import pickle
import random
import socket
import threading
import time

from thrift.transport import TTransport

from pinball.config.pinball_config import PinballConfig
from pinball.config.utils import get_log
from pinball.config.utils import get_unique_name

from pinball.master.snapshot import Snapshot
from pinball.master.thrift_lib.ttypes import ModifyRequest
from pinball.master.thrift_lib.ttypes import Query
from pinball.master.thrift_lib.ttypes import QueryAndOwnRequest
from pinball.master.thrift_lib.ttypes import QueryRequest
from pinball.master.thrift_lib.ttypes import Token
from pinball.master.thrift_lib.ttypes import TokenMasterException

from pinball.ui.data_builder import DataBuilder
from pinball.workflow.archiver import Archiver
from pinball.workflow.event import Event
from pinball.workflow.inspector import Inspector
from pinball.workflow.job_executor import JobExecutor
from pinball.workflow.name import Name
from pinball.workflow.signaller import Signal
from pinball.workflow.signaller import Signaller


__author__ = 'Pawel Garbacki'
__copyright__ = 'Copyright 2015, Pinterest, Inc.'
__credits__ = [__author__]
__license__ = 'Apache'
__version__ = '2.0'


LOG = get_log('pinball.workflow.worker')


class Worker(object):
    # Worker renews the ownership of the job token it owns every so often.
    _LEASE_TIME_SEC = 20 * 60  # 20 minutes

    # Delay between subsequent queries to the master.
    _INTER_QUERY_DELAY_SEC = 5

    def __init__(self, client, store, emailer):
        self._client = client
        self._emailer = emailer
        self._data_builder = DataBuilder(store)
        self._owned_job_token = None
        self._name = get_unique_name()
        self._inspector = Inspector(client)
        # The lock synchronizes access to shared attributes between the worker
        # thread and the lease renewer thread.
        self._lock = threading.Lock()
        self._lease_renewer = None
        self._executor = None
        self._test_only_end_if_no_runnable = False

    @staticmethod
    def _get_triggering_events(inputs):
        """Get a list of triggering events.

        Args:
            inputs: A list of lists where the elements of the outer list
                represent inputs of a job, while the elements of inner lists
                are names of events in those inputs.

        Returns:
            A list of event tokens, one per input, that may be used to trigger
            the job.  If any of the inputs has no events in it, the result list
            will be empty.

        Example:
            inputs = [[token('/workflows/wf/events/j/i1/e1'),
                       token('/workflows/wf/events/j/i1/e2')].
                      [token('/workflows/wf/events/j/i2/e3')]]
            return: [token('/workflows/wf/events/j/i1/e1'),
                     token('/workflows/wf/events/j/i2/e3')]

            inputs = [[token('/workflows/wf/events/j/i1/e1'),
                       token('/workflows/wf/events/j/i1/e2')].
                      []]
            return: []
        """
        triggering_events = []
        for events in inputs:
            if not events:
                return []
            triggering_events.append(events[0])
        return triggering_events

    def _move_job_token_to_runnable(self, job_token, triggering_event_tokens):
        """Move a job token to the runnable branch of the token tree.

        Token tree is the global, hierarchically structured token namespace.
        Args:
            job_token: The job token to make runnable.
            triggering_event_tokens: The list of events used to trigger the
                job.  These events will be removed from the master in the same
                call to that makes the job token runnable.
        Returns:
            True on success, otherwise False.
        """
        name = Name.from_job_token_name(job_token.name)
        name.job_state = Name.RUNNABLE_STATE
        job = pickle.loads(job_token.data)
        Worker._add_events_to_job(job, triggering_event_tokens)
        runnable_job_token = Token(name=name.get_job_token_name(),
                                   priority=job_token.priority,
                                   data=pickle.dumps(job))
        request = ModifyRequest(updates=[runnable_job_token],
                                deletes=triggering_event_tokens + [job_token])
        return self._send_request(request)

    @staticmethod
    def _add_events_to_job(job, triggering_event_tokens):
        """Put triggering events inside the job.

        Args:
            job: The job which should be augmented with the events.
            triggering_event_tokens: List of event tokens that triggered the
                job.
        """
        assert not job.events
        for event_token in triggering_event_tokens:
            if event_token.data:
                event = pickle.loads(event_token.data)
                # Optimization to make the job data structure smaller: do not
                # append events with no attributes.
                if event.attributes:
                    job.events.append(event)
            else:
                # This logic is here for backwards compatibility.
                # TODO(pawel): remove this logic after the transition to the
                # new model has been completed.
                name = Name.from_event_token_name(event_token.name)
                assert name.input == Name.WORKFLOW_START_INPUT

    def _make_job_runnable(self, job_token):
        """Attempt to make a job runnable.

        Query event tokens in job inputs.  If a combination of triggering
        events exist, remove those events and make the job runnable.
        Otherwise, do nothing.

        Args:
            job_token: The job token to make runnable.
        Returns:
            True if there were no errors during communication with the master,
            otherwise False.
        """
        job = pickle.loads(job_token.data)
        name = Name.from_job_token_name(job_token.name)
        request = QueryRequest(queries=[])
        # TODO(pawel): handle jobs with no dependencies
        assert job.inputs
        for input_name in job.inputs:
            prefix = Name()
            prefix.workflow = name.workflow
            prefix.instance = name.instance
            prefix.job = name.job
            prefix.input = input_name
            query = Query()
            query.namePrefix = prefix.get_input_prefix()
            query.maxTokens = 1
            request.queries.append(query)
        try:
            response = self._client.query(request)
        except TokenMasterException:
            # TODO(pawel): add a retry count and fail if a limit is reached.
            LOG.exception('error sending request %s', request)
            return False
        triggering_events = Worker._get_triggering_events(response.tokens)
        if triggering_events:
            return self._move_job_token_to_runnable(job_token,
                                                    triggering_events)
        return True

    def _make_runnable(self, workflow, instance):
        """Attempt to make jobs in a given workflow instance runnable.

        Go over all waiting jobs in a given workflow instance and try to make
        them runnable.

        Args:
            workflow: The name of the workflow whose jobs should be considered.
            instance: The workflow instance whose jobs should be considered.
        Returns:
            True if there were no errors during communication with the master,
            otherwise False.
        """
        name = Name()
        name.workflow = workflow
        name.instance = instance
        name.job_state = Name.WAITING_STATE
        query = Query(namePrefix=name.get_job_state_prefix())
        # TODO(pawel): to prevent multiple workers from trying to make the
        # same job runnable at the same time, this should be a
        # QueryAndOwnRequest.  Note that the current implementation is correct,
        # just inefficient.
        request = QueryRequest(queries=[query])
        try:
            response = self._client.query(request)
        except TokenMasterException:
            LOG.exception('error sending request %s', request)
            return False
        assert len(response.tokens) == 1
        for token in response.tokens[0]:
            if not self._make_job_runnable(token):
                return False
        return True

    def _has_no_runnable_jobs(self, workflow, instance):
        """Check if the workflow instance does not contain runnable jobs.

        Returns:
            True if we are certain that the workflow has no runnable jobs.
            Otherwise False.  If there were any errors during communication
            with the master, the return value is False.
        """
        name = Name(workflow=workflow,
                    instance=instance,
                    job_state=Name.RUNNABLE_STATE)
        query = Query(namePrefix=name.get_job_state_prefix())
        request = QueryRequest(queries=[query])
        try:
            response = self._client.query(request)
        except TokenMasterException:
            LOG.exception('error sending request %s', request)
            return False
        assert len(response.tokens) == 1
        if response.tokens[0]:
            return False
        return True

    def _is_done(self, workflow, instance):
        """Check if the workflow instance is done.

        A workflow is done if it does not have runnable jobs.

        Returns:
            True if we are certain that the workflow is not running.  Otherwise
            False.  If there were any errors during communication with the
            master, the return value is False.
        """
        # Attempt to make the workflow runnable and verify that no WAITING job
        # tokens were changed in the meantime.
        name = Name(workflow=workflow,
                    instance=instance,
                    job_state=Name.WAITING_STATE)
        query = Query(namePrefix=name.get_job_state_prefix())
        request = QueryRequest(queries=[query])
        try:
            snapshot = Snapshot(self._client, request)
        except:
            LOG.exception('error sending request %s', request)
            return False
        if not self._make_runnable(workflow, instance):
            return False
        if not self._has_no_runnable_jobs(workflow, instance):
            return False
        try:
            return not snapshot.refresh()
        except:
            LOG.exception('error sending request %s', request)
            return False

    def _process_signals(self, workflow, instance):
        """Process signals for a given workflow instance.

        Args:
            workflow: The workflow whose signals should be processed.
            instance: The instance whose signals should be processed.
        Returns:
            True if the worker should execute jobs in this instance.  Otherwise
            False.
        """
        signaller = Signaller(self._client, workflow, instance)
        archiver = Archiver(self._client, workflow, instance)
        if signaller.is_action_set(Signal.EXIT):
            return False
        if (signaller.is_action_set(Signal.ARCHIVE) and
                self._is_done(workflow, instance)):
            # TODO(pawel): enable this for all workflows after we gain
            # confidence that the master has enough memory to delay workflow
            # archiving.
            if workflow == 'indexing':
                ARCHIVE_DELAY_SEC = 7 * 24 * 60 * 60  # 7 days
            else:
                ARCHIVE_DELAY_SEC = 12 * 60 * 60  # 12 hours
            expiration_timestamp = int(time.time()) + ARCHIVE_DELAY_SEC
            if signaller.set_attribute_if_missing(Signal.ARCHIVE,
                                                  Signal.TIMESTAMP_ATTR,
                                                  expiration_timestamp):
                self._send_instance_end_email(workflow, instance)
            else:
                expiration_timestamp = signaller.get_attribute(
                    Signal.ARCHIVE, Signal.TIMESTAMP_ATTR)
                archiver.archive_if_expired(expiration_timestamp)
            return False
        if signaller.is_action_set(Signal.ABORT):
            if archiver.archive_if_aborted():
                self._send_instance_end_email(workflow, instance)
            return False
        if signaller.is_action_set(Signal.DRAIN):
            return False
        return True

    def _query_and_own_runnable_job_token(self, workflow, instance):
        """Attempt to own a runnable job token from a given workflow instance.

        Try to own a runnable job token in a given workflow instance.  The
        ownership of the qualifying job token lasts for a limited time so it
        has to be periodically renewed.

        Args:
            workflow: The name of the workflow whose jobs should be considered.
            instance: The workflow instance whose jobs should be considered.
        """
        assert not self._owned_job_token
        name = Name(workflow=workflow,
                    instance=instance,
                    job_state=Name.RUNNABLE_STATE)
        query = Query()
        query.namePrefix = name.get_job_state_prefix()
        query.maxTokens = 1
        request = QueryAndOwnRequest()
        request.query = query
        request.expirationTime = time.time() + Worker._LEASE_TIME_SEC
        request.owner = self._name
        try:
            response = self._client.query_and_own(request)
            if response.tokens:
                assert len(response.tokens) == 1
                self._owned_job_token = response.tokens[0]
        except TokenMasterException:
            LOG.exception('error sending request %s', request)

    def _own_runnable_job_token(self):
        """Attempt to own a runnable job token from any workflow."""
        assert not self._owned_job_token
        workflow_names = self._inspector.get_workflow_names()
        # Shuffle workflows to address starvation.
        random.shuffle(workflow_names)
        for workflow in workflow_names:
            instances = self._inspector.get_workflow_instances(workflow)
            time.sleep(Worker._INTER_QUERY_DELAY_SEC)
            random.shuffle(instances)
            for instance in instances:
                if self._process_signals(workflow, instance):
                    self._make_runnable(workflow, instance)
                    self._query_and_own_runnable_job_token(workflow, instance)
                    if self._owned_job_token:
                        return
            time.sleep(Worker._INTER_QUERY_DELAY_SEC)

    def _abort(self):
        """Abort the running job."""
        assert self._executor
        self._executor.abort()

    def _process_abort_signals(self):
        """Check if the running job should be aborted.

        Returns:
            False iff the job has been aborted.
        """
        name = Name.from_job_token_name(self._owned_job_token.name)
        abort = False
        try:
            signaller = Signaller(self._client, name.workflow, name.instance)
            abort = signaller.is_action_set(Signal.ABORT)
        except (TTransport.TTransportException, socket.timeout, socket.error):
            # We need this exception handler only in logic located in the
            # Timer thread.  If that thread fails, we should abort the process
            # and let the main thread decide what to do.
            LOG.exception('')
            abort = True
        if abort:
            self._abort()
        return not abort

    def _refresh_job_properties(self):
        """Record job properties in the master if they changed.

        If there are communication issues with the master, the running job
        gets aborted.

        Returns:
            False iff there was an error during communication with the master.
        """
        assert self._executor
        if self._executor.job_dirty:
            # The ordering here is important - we need to reset the changed
            # flag before updating the token.
            self._executor.job_dirty = False
            self._owned_job_token.data = pickle.dumps(self._executor.job)
            if not self._update_owned_job_token():
                self._abort()
                return False
        return True

    def _renew_ownership(self):
        """Periodic job token ownership renewal routine."""
        assert self._owned_job_token

        if not self._process_abort_signals():
            return

        if not self._refresh_job_properties():
            return

        now = time.time()
        if (self._owned_job_token.expirationTime <
                now + Worker._LEASE_TIME_SEC / 2):
            self._owned_job_token.expirationTime = (now +
                                                    Worker._LEASE_TIME_SEC)
            if not self._update_owned_job_token():
                self._abort()
                return

        with self._lock:
            if self._lease_renewer:
                self._lease_renewer = threading.Timer(
                    Worker._randomized_worker_polling_time(),
                    self._renew_ownership)
                self._lease_renewer.start()

    def _start_renew_ownership(self):
        """Start periodic renewal of the claimed job token ownership."""
        assert not self._lease_renewer
        self._lease_renewer = threading.Timer(
            Worker._randomized_worker_polling_time(),
            self._renew_ownership)
        self._lease_renewer.start()

    def _stop_renew_ownership(self):
        """Stop periodic renewal of the claimed job token ownership."""
        with self._lock:
            assert self._lease_renewer
            self._lease_renewer.cancel()
            lease_renewer = self._lease_renewer
            self._lease_renewer = None
        lease_renewer.join()

    def _send_request(self, request):
        """Send a modify request to the master.

        Args:
            request: The modify request to send.
        Returns:
            True on success, otherwise False.
        """
        try:
            self._client.modify(request)
            return True
        except TokenMasterException:
            LOG.exception('error sending request %s', request)
            return False

    def _get_output_event_tokens(self, job):
        """Create output event tokens for the owned job token.

        Args:
            job: The job which output tokens should be generated.
        Returns:
            A list of event tokens corresponding to the outputs of the owned
            job token.
        """
        assert self._owned_job_token
        job_name = Name.from_job_token_name(self._owned_job_token.name)
        output_name = Name()
        output_name.workflow = job_name.workflow
        output_name.instance = job_name.instance
        output_name.input = job_name.job
        event_tokens = []
        for output in job.outputs:
            output_name.job = output
            output_name.event = get_unique_name()
            event = Event(creator=self._name)
            assert job.history
            execution_record = job.history[-1]
            event.attributes = execution_record.get_event_attributes()
            event_tokens.append(Token(name=output_name.get_event_token_name(),
                                      data=pickle.dumps(event)))
        return event_tokens

    def _move_job_token_to_waiting(self, job, succeeded):
        """Move the owned job token to the waiting group.

        If the job succeeded, also post events to job outputs.  If the job
        failed or it is the final job (a job with no outputs),  post an archive
        signal to finish the workflow.

        Args:
            job: The job that should be stored in the data field of the waiting
                job token.
            succeeded: True if the job succeeded, otherwise False.
        """
        assert self._owned_job_token
        name = Name.from_job_token_name(self._owned_job_token.name)
        name.job_state = Name.WAITING_STATE
        waiting_job_token = Token(name=name.get_job_token_name(),
                                  priority=self._owned_job_token.priority,
                                  data=pickle.dumps(job))
        request = ModifyRequest(deletes=[self._owned_job_token],
                                updates=[waiting_job_token])
        if succeeded:
            request.updates.extend(self._get_output_event_tokens(job))
        if not job.outputs or not succeeded:
            # This is either the only job in the workflow with no outputs or a
            # failed job.  In either case, the workflow is done.
            signaller = Signaller(self._client,
                                  workflow=name.workflow,
                                  instance=name.instance)
            if not signaller.is_action_set(Signal.ARCHIVE):
                signal_name = Name(
                    workflow=name.workflow,
                    instance=name.instance,
                    signal=Signal.action_to_string(Signal.ARCHIVE))
                signal = Signal(Signal.ARCHIVE)
                signal_token = Token(name=signal_name.get_signal_token_name())
                signal_token.data = pickle.dumps(signal)
                request.updates.append(signal_token)
        self._send_request(request)

    def _unown(self, token):
        """Reset the ownership of a token.

        Args:
            token: The token whose ownership should be reset.
        """
        token.owner = None
        token.expirationTime = None

    def _keep_job_token_in_runnable(self, job):
        """Keep the owned job token in the runnable group.

        Refresh the job token data field with the provided job object, release
        the ownership of the token, and return it to the runnable group.

        Args:
            job: The job that should be stored in the data field of the job
                token.
        """
        assert self._owned_job_token
        request = ModifyRequest()
        self._owned_job_token.data = pickle.dumps(job)
        retry_delay_sec = job.retry_delay_sec
        if retry_delay_sec > 0:
            self._owned_job_token.expirationTime = (time.time() +
                                                    retry_delay_sec)
        else:
            self._unown(self._owned_job_token)
        request.updates = [self._owned_job_token]
        self._send_request(request)

    def _update_owned_job_token(self):
        """Update owned job token in the master.

        Returns:
            True if the update was successful, otherwise False.
        """
        assert self._owned_job_token
        request = ModifyRequest()
        request.updates = [self._owned_job_token]
        try:
            response = self._client.modify(request)
        except TokenMasterException:
            LOG.exception('error sending request %s', request)
            return False
        assert len(response.updates) == 1
        self._owned_job_token = response.updates[0]
        return True

    def _execute_job(self):
        """Execute the owned job."""
        assert self._owned_job_token
        job = pickle.loads(self._owned_job_token.data)
        name = Name.from_job_token_name(self._owned_job_token.name)
        self._executor = JobExecutor.from_job(name.workflow,
                                              name.instance,
                                              name.job,
                                              job,
                                              self._data_builder,
                                              self._emailer)
        success = self._executor.prepare()
        if success:
            self._owned_job_token.data = pickle.dumps(self._executor.job)
            success = self._update_owned_job_token()
            if success:
                self._start_renew_ownership()
                success = self._executor.execute()
                self._stop_renew_ownership()
        if success:
            self._move_job_token_to_waiting(self._executor.job, True)
        elif self._executor.job.retry():
            self._keep_job_token_in_runnable(self._executor.job)
        else:
            signaller = Signaller(self._client, name.workflow, name.instance)
            # If ARCHIVE is not set, this is the first failed job in the
            # workflow.
            first_failure = not signaller.is_action_set(Signal.ARCHIVE)
            self._move_job_token_to_waiting(self._executor.job, False)
            self._send_job_failure_emails(first_failure)
        self._executor = None
        self._owned_job_token = None
        # If needed, archive the workflow.
        self._process_signals(name.workflow, name.instance)

    def _send_instance_end_email(self, workflow, instance):
        try:
            schedule_data = self._data_builder.get_schedule(workflow)
            if not schedule_data:
                LOG.warning('no schedule found for workflow %s', workflow)
            elif schedule_data.emails:
                instance_data = self._data_builder.get_instance(workflow,
                                                                instance)
                jobs_data = self._data_builder.get_jobs(workflow, instance)
                self._emailer.send_instance_end_message(schedule_data.emails,
                                                        instance_data,
                                                        jobs_data)
        except:
            LOG.exception('error sending instance end email for workflow %s '
                          'instance %s', workflow, instance)

    def _send_job_failure_emails(self, first_failure):
        assert self._owned_job_token
        name = Name.from_job_token_name(self._owned_job_token.name)
        job = self._executor.job
        emails = set(job.emails)
        if first_failure:
            schedule_data = self._data_builder.get_schedule(name.workflow)
            if schedule_data:
                emails.update(schedule_data.emails)
            else:
                LOG.warning('no schedule found for workflow %s', name.workflow)
        if emails:
            execution = len(job.history) - 1
            job_execution_data = self._data_builder.get_execution(
                name.workflow, name.instance, name.job, execution)
            try:
                self._emailer.send_job_execution_end_message(
                    list(emails), job_execution_data)
            except:
                LOG.exception('error sending job failure email for '
                              'workflow %s instance %s job %s execution %d',
                              name.workflow,
                              name.instance,
                              name.job,
                              execution)

    @staticmethod
    def _randomized_worker_polling_time():
        """Generate random worker polling time."""
        return (1.0 + random.random()) * PinballConfig.WORKER_POLL_TIME_SEC

    def run(self):
        """Run the worker."""
        LOG.info('Running worker ' + self._name)
        while True:
            signaller = Signaller(self._client)
            if signaller.is_action_set(Signal.EXIT):
                return
            if not signaller.is_action_set(Signal.DRAIN):
                self._own_runnable_job_token()
            if self._owned_job_token:
                self._execute_job()
            elif self._test_only_end_if_no_runnable:
                return
            else:
                time.sleep(Worker._randomized_worker_polling_time())
        LOG.info('Exiting worker ' + self._name)



# Copyright 2015, Pinterest, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utilities for executing shell commands.

Anything related to running subprocesses, or interacting with the shell or
prompt should go here.
"""

import subprocess

from pinball_ext.common.decorators import retry
from pinball_ext.common.utils import get_logger


__author__ = 'Dmitry Chechik, Changshu Liu'
__copyright__ = 'Copyright 2015, Pinterest, Inc.'
__credits__ = [__author__]
__license__ = 'Apache'
__version__ = '2.0'


LOG = get_logger("pinball_ext.common.shell_utils")

_DEFAULT_SSH_OPTS = [
    '-o', 'UserKnownHostsFile=/dev/null',
    '-o', 'StrictHostKeyChecking=no',
    # Suppress warnings while SSHing.
    '-o', 'LogLevel=quiet',
]

_DEFAULT_RSYNC_OPTS = [
    '-u',  # Skip files that are newer on the receiver.
    '-L',  # Transform symlinks to real files.
    '-v',
]


@retry(subprocess.CalledProcessError, tries=3, delay=1)
def scp_file_to_host(user_name,
                     remote_host,
                     local_path,
                     remote_path,
                     ssh_options=_DEFAULT_SSH_OPTS,
                     ssh_port=22,
                     ssh_key_file=None):
    """SCP the local_path file to the remote host.

    Args:
        ``user_name`` - username on remote machine.
        ``host`` - remote host.
        ``local_path`` - path on local machine.
        ``remote_path`` - path on remote machine.
        ``options`` - any additional options.
        ``port`` - port to connect to.
        ``key`` - key file to use for connection, if any.
    Raises:
        CalledProcessError if scp fails.
    """
    scp_command = ['scp'] + ssh_options + ['-P', str(ssh_port)]
    if ssh_key_file:
        scp_command += ['-i', ssh_key_file]
    scp_command += [local_path]
    scp_command += ['%s@%s:%s' % (user_name, remote_host, remote_path)]
    scp_command = ' '.join(scp_command)

    LOG.info('Running command: %s', scp_command)

    subprocess.check_call(scp_command, shell=True)


@retry(subprocess.CalledProcessError, tries=3, delay=1)
def rsync_over_ssh(user_name,
                   remote_host,
                   local_path,
                   remote_path,
                   rsync_options=_DEFAULT_RSYNC_OPTS,
                   ssh_options=_DEFAULT_SSH_OPTS,
                   ssh_port=22,
                   ssh_key_file=None):
    """rsync the local_path to remote host over SSH.

    Args:
        username - username on remote machine.
        host - remote host.
        local_path - path on local machine.
        remote_path - path on remote machine.
        rsync_options - rsync options.
        ssh_options - extra ssh options.
        ssh_port - port to connect to.
        ssh_key_file - key file to use for connection, if any.

    Raises:
        CalledProcessError if scp fails.
    """
    ssh_command = ['ssh'] + ssh_options + ['-p', str(ssh_port)]
    if ssh_key_file:
        ssh_command += ['-i', ssh_key_file]

    command = ['rsync'] + rsync_options + ['-e ' + '\"%s\"' % ' '.join(ssh_command)]
    command += [local_path]
    command += ['%s@%s:%s' % (user_name, remote_host, remote_path)]
    command = ' '.join(command)

    LOG.info('Running command:%s' % command)

    subprocess.check_call(command, shell=True)


def run_ssh_command(user_name,
                    host_name,
                    command,
                    cmd_work_dir,
                    ssh_options=_DEFAULT_SSH_OPTS,
                    ssh_port=22,
                    ssh_key_file=None,
                    check_output=False,
                    forward_agent=False):
    """Run the given SSH command and returns the stdout response.

    Args:
        host - host to SSH to.
        command - The command to run.
        cmd_work_dir - Path on remote host to cd to before running the command.
        username - username to SSH as.
        ssh_options - list of additional options to pass to the SSH command.
        ssh_port - port to SSH to.
        ssh_key_file - Path to a key file to use to SSH.
        check_output - Check that the command ran successfully?
                       If so, this will return nothing and raise a
                       CalledProcessError on error.
                       If not, it will return a file-like object with the
                       command stdout to read.

    Returns:
        A file pointer to the stdout output of the process.
    """
    ssh_command = ['ssh'] + ssh_options + ['-p', str(ssh_port)]
    if ssh_key_file:
        ssh_command += ['-i', ssh_key_file]

    if forward_agent:
        ssh_command += ['-A']

    ssh_command += ['%s@%s' % (user_name, host_name)]
    ssh_command += ["cd %s && %s" % (cmd_work_dir, command)]

    LOG.info('Running command:%s' % ssh_command)

    if check_output:
        return subprocess.check_call(ssh_command)
    else:
        ssh = subprocess.Popen(ssh_command, stdout=subprocess.PIPE)

        return StdoutProcessWrapper(ssh_command, ssh)


class StdoutProcessWrapper(object):
    """A file-like object that wraps around the subprocess.POpen call.

    It acts as a file that will read the data from the popen's stdout,
    and will verify and raise a CalledProcessError if the underlying
    process errors out at any point.

    We can't just return POpen.stdout from these commands, because if
    we do, it won't check if there process exists with a bad return code
    when we read from it.
    """
    def __init__(self, cmd, popen_obj):
        self.popen_obj = popen_obj
        self.cmd = cmd

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        # If an exception happened in the body of the with we want to
        # return False to let the exception propagate.
        if exc_type:
            return False

        # Otherwise check that SSH completed succesfully.
        self._verify_success()
        return False

    def _verify_success(self):
        returncode = self.popen_obj.wait()
        if returncode:
            raise subprocess.CalledProcessError(returncode, self.cmd)

    def close(self):
        self._verify_success()

    def next(self):
        try:
            return self.popen_obj.stdout.next()
        except StopIteration as e:
            self.popen_obj.wait()
            self._verify_success()
            raise e

    def __iter__(self):
        while True:
            yield self.next()

    def read(self, bytes_to_read=None):
        if bytes_to_read:
            # Only read bytes_to_read read bytes.
            result = self.popen_obj.stdout.read(bytes_to_read)
            return result
        else:
            # Block until we read everything.
            result = self.popen_obj.stdout.read()
            self._verify_success()
            return result

    def readline(self):
        result = self.popen_obj.stdout.readline()
        if not result:
            self._verify_success()
        return result



# Copyright 2015, Pinterest, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Populate the database with some tokens.  Useful for testing."""
import argparse
import os
import pickle
import sys
import time

from pinball.master.blessed_version import BlessedVersion
from pinball.master.master_handler import MasterHandler
from pinball.master.thrift_lib.ttypes import Token
from pinball.persistence.store import DbStore
from pinball.scheduler.schedule import WorkflowSchedule
from pinball.workflow.job import ShellJob
from pinball.workflow.job_executor import ExecutionRecord
from pinball.workflow.name import Name
from pinball.workflow.signaller import Signal


__author__ = 'Pawel Garbacki'
__copyright__ = 'Copyright 2015, Pinterest, Inc.'
__credits__ = [__author__]
__license__ = 'Apache'
__version__ = '2.0'


def _try_generate_blessed_version(store):
    blessed_version = BlessedVersion(MasterHandler._BLESSED_VERSION,
                                     MasterHandler._MASTER_OWNER)
    active_tokens = store.read_active_tokens()
    for token in active_tokens:
        if token.name == blessed_version.name:
            return
    store.commit_tokens([blessed_version])


def _generate_job_token(workflow, instance, job, executions, max_jobs):
    if job == 0:
        inputs = [Name.WORKFLOW_START_INPUT]
    else:
        inputs = ['job_%d' % (job - 1)]
    if job == max_jobs - 1:
        outputs = []
    else:
        outputs = ['job_%d' % (job + 1)]
    shell_job = ShellJob(name='job_%d' % job, inputs=inputs, outputs=outputs,
                         command='some command %d' % job)
    for e in range(0, executions):
        start_time = 1000000 * workflow + 10000 * instance + 100 * job + e + 1
        end_time = start_time + 10 * e + 1
        DIR = '/tmp/pinball/logs'
        if not os.path.exists(DIR):
            os.makedirs(DIR)
        LOG_PATTERN = '%s/%%s.%%d.%%s' % DIR
        info_log_file = LOG_PATTERN % (job, start_time, 'info')
        with open(info_log_file, 'w') as f:
            f.write('some info log of execution %d' % e)
        error_log_file = LOG_PATTERN % (job, start_time, 'error')
        with open(error_log_file, 'w') as f:
            f.write('some error log of execution %d' % e)
        record = ExecutionRecord(
            info='some_command %d some_args %d' % (e, e),
            instance='instance_%d' % instance,
            start_time=start_time,
            end_time=end_time,
            exit_code=(workflow + instance + e) % 2,
            logs={'info': info_log_file, 'error': error_log_file})
        shell_job.history.append(record)
    name = Name(workflow='workflow_%d' % workflow,
                instance='instance_%d' % instance,
                job_state=Name.WAITING_STATE,
                job='job_%d' % job)
    return Token(name=name.get_job_token_name(),
                 version=1000000 * workflow + 10000 * instance + 100 * job,
                 priority=job,
                 data=pickle.dumps(shell_job))


def _generate_workflows_tokens(offset, workflows, instances, jobs, executions):
    result = []
    for w in range(offset, offset + workflows):
        for i in range(0, instances):
            for j in range(0, jobs):
                result.append(_generate_job_token(w, i, j, executions, jobs))
    return result


def _generate_schedule_tokens(workflows):
    result = []
    for w in range(workflows):
        next_run_time = time.time() + (365 + w) * 24 * 60 * 60
        recurrence = min(365 * 24 * 60 * 60, 60 ** w)
        workflow = 'workflow_%d' % w
        schedule = WorkflowSchedule(next_run_time,
                                    recurrence_seconds=recurrence,
                                    overrun_policy=w % 4, workflow=workflow)
        name = Name(workflow=workflow)
        result.append(Token(name=name.get_workflow_schedule_token_name(),
                            version=100000000 * w,
                            owner='some_owner',
                            expirationTime=next_run_time,
                            data=pickle.dumps(schedule)))
    return result


def _generate_signal_tokens(workflows):
    result = []
    for w in range(0, workflows, 2):
        workflow = 'workflow_%d' % w
        signal = Signal(Signal.DRAIN)
        name = Name(workflow=workflow,
                    signal=Signal.action_to_string(signal.action))
        result.append(Token(name=name.get_signal_token_name(),
                            version=10000000000 * w,
                            data=pickle.dumps(signal)))
    return result


def generate_workflows(active_workflows, archived_workflows, instances, jobs,
                       executions, store):
    _try_generate_blessed_version(store)
    active_tokens = _generate_workflows_tokens(0,
                                               active_workflows,
                                               instances,
                                               jobs,
                                               executions)

    archived_tokens = _generate_workflows_tokens(active_workflows,
                                                 archived_workflows,
                                                 instances,
                                                 jobs,
                                                 executions)

    schedule_tokens = _generate_schedule_tokens(active_workflows +
                                                archived_workflows)

    signal_tokens = _generate_signal_tokens(active_workflows)

    store.commit_tokens(active_tokens + archived_tokens + schedule_tokens +
                        signal_tokens)
    store.archive_tokens(archived_tokens)


def main():
    parser = argparse.ArgumentParser(
        description='Start Pinball master server.')
    parser.add_argument('-w',
                        '--active_workflows',
                        dest='active_workflows',
                        type=int,
                        default=5,
                        help='number of active workflows to generate')
    parser.add_argument('-r',
                        '--archived_workflows',
                        dest='archived_workflows',
                        type=int,
                        default=5,
                        help='number of archived workflows to generate')
    parser.add_argument('-i',
                        '--instances',
                        dest='instances',
                        type=int,
                        default=5,
                        help='number of instances per workflow')
    parser.add_argument('-j',
                        '--jobs',
                        dest='jobs',
                        type=int,
                        default=5,
                        help='number of jobs per workflow')
    parser.add_argument('-e',
                        '--executions',
                        dest='executions',
                        type=int,
                        default=5,
                        help='number of executions per job')
    options = parser.parse_args(sys.argv[1:])

    generate_workflows(options.active_workflows,
                       options.archived_workflows,
                       options.instances,
                       options.jobs,
                       options.executions,
                       DbStore())


if __name__ == '__main__':
    main()



# Copyright 2015, Pinterest, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Web request handlers."""
import collections
import httplib2
import json
import socket
import traceback
import time

from django.contrib import messages
from django.http import HttpResponse
from django.http import HttpResponseRedirect
from django.http import HttpResponseServerError
from django.shortcuts import render_to_response
from django.template.context import RequestContext
from django.views.generic.base import TemplateView

from pinball.authentication import oauth2
from pinball.config.pinball_config import PinballConfig
from pinball.config.utils import get_log
from pinball.tools.workflow_util import run_command
from pinball.ui.cache_thread import get_workflows_json
from pinball.ui.data_builder import DataBuilder
from pinball.ui.utils import get_workflow_jobs_from_parser
from pinball.ui.workflow_graph import WorkflowGraph
from pinball.persistence.store import DbStore
from pinball.workflow.signaller import Signal


__author__ = 'Pawel Garbacki'
__copyright__ = 'Copyright 2015, Pinterest, Inc.'
__credits__ = [__author__]
__license__ = 'Apache'
__version__ = '2.0'


LOG = get_log('pinball.ui.views')
# Custom message level
SIGNIN = 35


def _serialize(elements):
    elements_list = []
    for element in elements:
        elements_list.append(element.format())
    to_serialize = {'aaData': elements_list}
    return json.dumps(to_serialize)


def workflows(_):
    try:
        workflows_json = get_workflows_json()
    except:
        LOG.exception('')
        return HttpResponseServerError(traceback.format_exc())
    else:
        return HttpResponse(workflows_json, mimetype='application/json')


def instances(request):
    try:
        workflow = request.GET['workflow']
        data_builder = DataBuilder(DbStore(), use_cache=True)
        instances_data = data_builder.get_instances(workflow)
        instances_json = _serialize(instances_data)
    except:
        LOG.exception('')
        return HttpResponseServerError(traceback.format_exc())
    else:
        return HttpResponse(instances_json, mimetype='application/json')


def jobs(request):
    try:
        data_builder = DataBuilder(DbStore(), use_cache=True)
        workflow = request.GET['workflow']
        instance = request.GET['instance']
        if instance == 'latest':
            instance = data_builder.get_latest_instance(workflow).instance
        jobs_data = data_builder.get_jobs(workflow, instance)
        jobs_json = _serialize(jobs_data)
    except:
        LOG.exception('')
        return HttpResponseServerError(traceback.format_exc())
    else:
        return HttpResponse(jobs_json, mimetype='application/json')


def graph(request):
    try:
        data_builder = DataBuilder(DbStore(), use_cache=True)
        workflow = request.GET['workflow']
        if 'instance' in request.GET:
            instance = request.GET['instance']
            if instance == 'latest':
                instance = data_builder.get_latest_instance(workflow).instance
            jobs_data = data_builder.get_jobs(workflow=workflow,
                                              instance=instance)
            instance_data = data_builder.get_instance(workflow=workflow,
                                                      instance=instance)
            workflow_graph = WorkflowGraph(jobs_data, instance_data)
        else:
            workflow_graph = WorkflowGraph.from_parser(workflow)
    except:
        LOG.exception('')
        return HttpResponseServerError(traceback.format_exc())
    else:
        return HttpResponse(workflow_graph.get_svg(), mimetype='image/svg+xml')


def executions(request):
    try:
        workflow = request.GET['workflow']
        instance = request.GET.get('instance')
        job = request.GET['job']
        data_builder = DataBuilder(DbStore())
        if instance:
            executions_data = data_builder.get_executions(workflow,
                                                          instance,
                                                          job)
        else:
            executions_data = data_builder.get_executions_across_instances(
                workflow, job)
        executions_json = _serialize(executions_data)
    except:
        LOG.exception('')
        return HttpResponseServerError(traceback.format_exc())
    else:
        return HttpResponse(executions_json, mimetype='application/json')


class ExecutionView(TemplateView):
    template_name = 'execution.html'

    def get_context_data(self, **kwargs):
        context = super(ExecutionView, self).get_context_data(**kwargs)
        workflow = self.request.GET['workflow']
        instance = self.request.GET['instance']
        job = self.request.GET['job']
        execution = int(self.request.GET['execution'])
        data_builder = DataBuilder(DbStore())
        execution_data = data_builder.get_execution(workflow,
                                                    instance,
                                                    job,
                                                    execution)
        formatted_data = execution_data.format()
        for key, value in formatted_data.items():
            context[key] = value
        properties = []
        for key, value in execution_data.properties.items():
            properties.append('%s=%s' % (key, value))
        context['properties'] = ', '.join(properties)
        if not execution_data.end_time:
            context['end_time'] = ''
        if execution_data.exit_code is None:
            context['exit_code'] = ''
        return context


def file_content(request):
    try:
        workflow = request.GET['workflow']
        instance = request.GET['instance']
        job = request.GET['job']
        execution = int(request.GET['execution'])
        log_type = request.GET['log_type']
        if execution < 0:
            return HttpResponseServerError(
                'execution must not be negative; got ' + execution)
        data_builder = DataBuilder(DbStore())
        file_data = data_builder.get_file_content(workflow, instance, job,
                                                  execution, log_type)
    except:
        LOG.exception('')
        return HttpResponseServerError(traceback.format_exc())
    else:
        return HttpResponse(file_data, mimetype='text/plain')


def schedules(_):
    try:
        data_builder = DataBuilder(DbStore())
        schedules_data = data_builder.get_schedules()
        schedules_json = _serialize(schedules_data)
    except:
        LOG.exception('')
        return HttpResponseServerError(traceback.format_exc())
    else:
        return HttpResponse(schedules_json, mimetype='application/json')


class ScheduleView(TemplateView):
    template_name = 'schedule.html'

    def get_context_data(self, **kwargs):
        context = super(ScheduleView, self).get_context_data(**kwargs)
        workflow = self.request.GET['workflow']
        data_builder = DataBuilder(DbStore())
        schedule_data = data_builder.get_schedule(workflow)
        formatted_schedule = schedule_data.format()
        for key, value in formatted_schedule.items():
            context[key] = value
        context['emails'] = ' '.join(schedule_data.emails)
        return context


def jobs_from_config(request):
    try:
        workflow = request.GET['workflow']
        jobs_data = get_workflow_jobs_from_parser(workflow)
        jobs_json = _serialize(jobs_data)
    except:
        LOG.exception('')
        return HttpResponseServerError(traceback.format_exc())
    else:
        return HttpResponse(jobs_json, mimetype='application/json')


def command(request):
    try:
        args = {}
        for key in request.GET:
            args[key] = request.GET[key]
        args['force'] = True
        if 'workflow' not in args:
            args['workflow'] = None
        if 'instance' not in args:
            args['instance'] = None
        Options = collections.namedtuple('Options', args.keys())
        options = Options(**args)
        output = run_command(options)
    except:
        LOG.exception('')
        return HttpResponseServerError(traceback.format_exc())
    else:
        return HttpResponse(json.dumps(output), mimetype='application/json')


class TokenPathsView(TemplateView):
    template_name = 'token_paths.html'

    def get_context_data(self, **kwargs):
        context = super(TokenPathsView, self).get_context_data(**kwargs)
        path = self.request.GET['path']
        if not path or not path.startswith('/'):
            return context
        path_elements = path.split('/')[1:]
        if path.endswith('/'):
            path_elements.pop()
        if not path_elements:
            context['basename'] = ''
        else:
            context['basename'] = path_elements[-1]
        # A list of tuples (parent_name, parent_prefix).
        parents = []
        prefix = '/'
        for element in path_elements[:-1]:
            prefix += '%s/' % element
            parents.append((element, prefix))
        context['parents'] = parents
        return context


def token_paths(request):
    try:
        path = request.GET['path']
        data_builder = DataBuilder(DbStore())
        tokens_data = data_builder.get_token_paths(path)
        tokens_json = _serialize(tokens_data)
    except:
        LOG.exception('')
        return HttpResponseServerError(traceback.format_exc())
    else:
        return HttpResponse(tokens_json, mimetype='application/json')


class TokenView(TokenPathsView):
    template_name = 'token.html'

    def get_context_data(self, **kwargs):
        context = super(TokenView, self).get_context_data(**kwargs)
        token_name = self.request.GET['path']
        data_builder = DataBuilder(DbStore())
        token_data = data_builder.get_token(token_name)
        token_format = token_data.format()
        for key, value in token_format.items():
            context[key] = value
        return context


def _is_master_alive():
    try:
        s = socket.socket()
        host = 'localhost'
        s.connect((host, PinballConfig.MASTER_PORT))
        s.close()
    except:
        return False
    return True


def status(request):
    try:
        workflow = request.GET.get('workflow')
        instance = request.GET.get('instance')
        data_builder = DataBuilder(DbStore())
        status = []
        if data_builder.is_signal_set(workflow, instance, Signal.EXIT):
            status = ['exiting']
        elif data_builder.is_signal_set(workflow, instance, Signal.ABORT):
            status = ['aborting']
        elif data_builder.is_signal_set(workflow, instance, Signal.DRAIN):
            status = ['draining']
        if not _is_master_alive():
            status.append('no master at %s:%d' % (socket.gethostname(),
                                                  PinballConfig.MASTER_PORT))
        status_json = json.dumps(status)
    except:
        LOG.exception('')
        return HttpResponseServerError(traceback.format_exc())
    else:
        return HttpResponse(status_json, mimetype='application/json')


def signin(request):
    oauth2_flow = oauth2.OAuth2Flow()
    context = {'domains': oauth2_flow.get_domains(), 'STATIC_URL': PinballConfig.STATIC_URL}
    if request.method == 'POST' and 'signin-domain' in request.POST.keys():
        domain = request.POST.get('signin-domain')
        if not oauth2_flow.domain_authenticated(domain):
            messages.add_message(request, SIGNIN, 'Domain not authorized: %s.' % domain,
                                 fail_silently=True)
            return render_to_response('signin.html', context,
                                      context_instance=RequestContext(request),
                                      mimetype='text/html')
        else:
            flow = oauth2_flow.get_flow(domain)
            auth_uri = flow.step1_get_authorize_url()
            return HttpResponseRedirect(auth_uri)
    else:
        return render_to_response('signin.html', context,
                                  context_instance=RequestContext(request),
                                  mimetype='text/html')


def auth_return(request):
    oauth2_flow = oauth2.OAuth2Flow()
    domains = oauth2_flow.get_domains()
    flow = oauth2.OAuth2Flow().get_flow(domains[1])
    # disable SSL certificate validation for exchanging access code
    http = httplib2.Http()
    http.disable_ssl_certificate_validation = True
    credential = flow.step2_exchange(request.GET.get('code'), http)
    credential_token = json.loads(credential.to_json())['id_token']
    if credential_token['email_verified'] and credential_token['hd'] in domains:
        email = credential_token['email']
        crypter = oauth2.Crypter()
        encrypted_email = crypter.encrypt(email)
        encrypted_domain = crypter.encrypt(credential_token['hd'])
        encrypted_token = crypter.encrypt(credential.access_token)
        response = HttpResponseRedirect('/')
        # cookie expires after a week
        response.set_cookie('login', encrypted_email, max_age=7 * 24 * 60 * 60)
        response.set_cookie('domain_url', encrypted_domain, max_age=7 * 24 * 60 * 60)
        response.set_cookie('user_id', email, max_age=7 * 24 * 60 * 60)
        response.set_cookie('token', encrypted_token)
        return response
    else:
        messages.add_message(request, SIGNIN, 'Authentication failed.')
        response = HttpResponseRedirect('/logout/')


def logout(request):
    credential_token = request.COOKIES.get('token', '')
    if credential_token == '':
        messages.add_message(request, SIGNIN, 'Logged out successfully.',
                             fail_silently=True)
        return HttpResponseRedirect('/signin/')
    crypter = oauth2.Crypter()
    try:
        logout_uri = 'https://accounts.google.com/o/oauth2/revoke?token=%s' \
            % crypter.decrypt(credential_token)
    except oauth2.CryptoException:
        response = HttpResponseRedirect('/signin/')
    else:
        http = httplib2.Http()
        http.disable_ssl_certificate_validation = True
        resp = http.request(logout_uri, 'GET')

        response = HttpResponseRedirect('/signin/')
        # Need to wait for Google to process the revoke request
        if resp[0].status == 200:
            time.sleep(2)
            response = HttpResponseRedirect('/logout/')

    response.set_cookie('user_id', '')
    response.set_cookie('login', '')
    response.set_cookie('domain_url', '')
    response.set_cookie('token', '')
    return response



# Copyright 2015, Pinterest, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Various misc utilities"""

import logging
import random
import string


__author__ = 'Changshu Liu'
__copyright__ = 'Copyright 2015, Pinterest, Inc.'
__credits__ = [__author__]
__license__ = 'Apache'
__version__ = '2.0'


def get_random_string(size=10, chars=string.ascii_lowercase + string.digits):
    """Get a random string of given size and composed of given characters.

    Args:
        size: Size of random string generated.
        chars: Set of characters which make up the random string.
    """
    return ''.join(random.choice(chars) for x in range(size))


def get_logger(module):
    logging.basicConfig(
        format='[%(asctime)s] - %(name)s - %(levelname)-8s"%(message)s"',
        datefmt='%Y-%m-%d %a %H:%M:%S')
    log = logging.getLogger(module)
    log.setLevel(logging.INFO)
    return log


def parse_arguments(args):
    """Parse argument text into Python dictionary.

    Args:
        args: a list of arguments separated by ','. Each argument could be
            key/value pair in format: key1=value1 or it could just be a
            flag_name.

    Returns:
        Corresponding Python dict that maps:
            key1->value1,
            flag_name->True
    """
    arg_dict = {}
    if args:
        for arg in args.split(','):
            if '=' in arg:
                # The key from foo=bar is foo. For foo=bar=xyz, the key is foo
                # and the value is bar=xyz.
                key, value = arg.split('=', 1)
            else:
                # Just some existence flag. Assume it is True.
                key = arg
                value = True
            arg_dict[key.strip()] = value
    return arg_dict



# Copyright 2015, Pinterest, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Job template translates a user defined job into Pinball job that will be
executed by Pinball workers directly.

Condition is a special type of job.
"""

import abc
from datetime import datetime, timedelta

from pinball.workflow.job import ShellConditionJob
from pinball.workflow.job import ShellJob


__author__ = 'Pawel Garbacki, Changshu Liu, Mao Ye'
__copyright__ = 'Copyright 2015, Pinterest, Inc.'
__credits__ = [__author__]
__license__ = 'Apache'
__version__ = '2.0'


class JobTemplateBase(object):
    """An interface for job templates."""
    __metaclass__ = abc.ABCMeta

    def __init__(self, name, write_lock=None, max_attempts=1, emails=None,
                 priority=None, warn_timeout_sec=None, abort_timeout_sec=None):
        self.name = name
        self.write_lock = write_lock
        self._max_attempts = max_attempts
        self._emails = emails if emails is not None else []
        self.priority = priority
        self._warn_timeout_sec = warn_timeout_sec
        self._abort_timeout_sec = abort_timeout_sec

    def __eq__(self, other):
        return self.name == other.name

    def __hash__(self):
        return ('name=%s' % self.name).__hash__()

    @abc.abstractmethod
    def get_pinball_job(self, inputs, outputs, params):
        """Get a Pinball job from user defined job.

        Args:
            inputs: The list of job inputs.
            outputs: The list of job outputs.
            params: The dictionary with extra parameters.
        Returns:
            The pinball job specializing the template.
        """
        return


class JobTemplate(JobTemplateBase):
    """A template that supports running user defined jobs.

    All user defined jobs are supposed to be derived from:
        pinball_ext.job.basic_jobs.JobBase

    Constructor args:
        name: Python class name of the user defined job.
        executor: name of the executor. Available executors can be found from
            pinball_ext.executor.common.Platform
        executor_config: dict config for a specific executor.
    """
    command_template = (
        'cd %(job_repo_dir)s && python -m pinball_ext.job.job_runner '
        '--job_import_dirs_config=%(job_import_dirs_config)s '
        '--job_class_name=%(job_class_name)s '
        '--job_params="%(job_params)s" '
        '%(executor_params)s'
        '%(executor_config)s'
        '%(runner_extra_params)s')

    def __init__(self, name, executor=None, executor_config=None, write_lock=None,
                 max_attempts=1, emails=None, priority=None,
                 warn_timeout_sec=None, abort_timeout_sec=None):
        assert name, 'name should be set.'

        super(JobTemplate, self).__init__(
            name=name,
            write_lock=write_lock,
            max_attempts=max_attempts,
            emails=emails,
            priority=priority,
            warn_timeout_sec=warn_timeout_sec,
            abort_timeout_sec=abort_timeout_sec)

        self._job_class_name = name
        self._executor = executor
        self._executor_config = executor_config

    def get_pinball_job(self, inputs, outputs, params):
        assert 'job_repo_dir' in params and 'job_import_dirs_config' in params

        # Config job_params
        job_params = params.get('job_params', {})
        # TODO(csliu): end_date should always be passed from params
        if 'end_date' not in params:
            yesterday = str(datetime.utcnow().date() - timedelta(days=1))
            job_params['end_date'] = yesterday
        else:
            job_params['end_date'] = params['end_date']
        assert job_params, 'job_params should not be empty!'
        job_params_text = ','.join(
            ['%s=%s' % (k, v) for k, v in job_params.iteritems()])

        # Config executor for job runner. (Only ClusterJob needs executor)
        executor_params = ''
        if self._executor:
            executor_params = '--executor=%s ' % self._executor

        # Executor config.
        executor_config = ''
        if self._executor_config:
            executor_config_text = ','.join(
                ['%s=%s' % (k, v) for k, v in self._executor_config.iteritems()])
            executor_config = '--executor_config="%s" ' % executor_config_text

        runner_extra_params = ''
        if self.write_lock:
            runner_extra_params += '--write_lock=%s ' % self.write_lock

        # Construct job runner command line.
        job_runner_command = self.command_template % {
            'job_repo_dir': params['job_repo_dir'],
            'job_import_dirs_config': params['job_import_dirs_config'],
            'job_class_name': self._job_class_name,
            'job_params': job_params_text,
            'executor_params': executor_params,
            'executor_config': executor_config,
            'runner_extra_params': runner_extra_params
        }

        return ShellJob(name=self.name, inputs=inputs, outputs=outputs,
                        command=job_runner_command,
                        emails=self._emails, max_attempts=self._max_attempts,
                        warn_timeout_sec=self._warn_timeout_sec,
                        abort_timeout_sec=self._abort_timeout_sec)


class CommandJobTemplate(JobTemplateBase):
    """The template to invoke a command-line job."""
    def __init__(self, name, command, write_lock=None,
                 max_attempts=None, emails=None, priority=None,
                 warn_timeout_sec=None, abort_timeout_sec=None):
        super(CommandJobTemplate, self).__init__(name,
                                                 write_lock,
                                                 max_attempts,
                                                 emails,
                                                 priority,
                                                 warn_timeout_sec,
                                                 abort_timeout_sec)
        self._command = command

    def get_pinball_job(self, inputs, outputs, params=None):
        params = params if params else {}
        command = self._command % params
        max_attempts = (self._max_attempts if self._max_attempts is not None
                        else 1)
        return ShellJob(name=self.name, inputs=inputs, outputs=outputs,
                        emails=self._emails, max_attempts=max_attempts,
                        warn_timeout_sec=self._warn_timeout_sec,
                        abort_timeout_sec=self._abort_timeout_sec,
                        command=command)


class ConditionTemplateBase(object):
    """An interface for condition templates."""
    __metaclass__ = abc.ABCMeta

    def __init__(self, name,
                 max_attempts=10,
                 retry_delay_sec=5 * 60,
                 emails=None,
                 priority=None):
        self.name = name
        self.max_attempts = max_attempts
        self.retry_delay_sec = retry_delay_sec
        self.emails = emails if emails is not None else []
        self.priority = priority

    def __eq__(self, other):
        return self.name == other.name

    def __hash__(self):
        return ('name=%s' % self.name).__hash__()

    @abc.abstractmethod
    def get_pinball_condition(self, outputs):
        """Create a pinball condition specification.

        Args:
            outputs: The list of condition outputs.
        Returns:
            The pinball condition specializing the template.
        """
        return


class CommandConditionTemplate(ConditionTemplateBase):
    """The template to run a custom command."""
    # 72 * 20 * 60 = seconds in a day.
    def __init__(self, name,
                 max_attempts=72,
                 retry_delay_sec=20 * 60,
                 emails=None,
                 priority=None):
        super(CommandConditionTemplate, self).__init__(
            name,
            max_attempts=max_attempts,
            retry_delay_sec=retry_delay_sec,
            emails=emails,
            priority=priority)

    def get_pinball_condition(self, outputs, params=None):
        command = params['command']
        cleanup_template = params.get('cleanup_template', None)
        return ShellConditionJob(name=self.name, outputs=outputs,
                                 emails=self.emails,
                                 max_attempts=self.max_attempts,
                                 retry_delay_sec=self.retry_delay_sec,
                                 command=command,
                                 cleanup_template=cleanup_template)



# Copyright 2015, Pinterest, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Defines example jobs."""

import datetime
import os

from pinball_ext.job.basic_jobs import PythonJob
from pinball_ext.job.basic_jobs import CommandLineJob
from pinball_ext.job_templates import CommandJobTemplate


JOB_IMPORT_DIRS = [os.path.dirname(os.path.realpath(__file__))]


class ExamplePythonJob(PythonJob):

    def _setup(self):
        print 'Do some setup in example python job!'

    def _execute(self):
        print 'Current time is %s' % str(datetime.datetime.now())


class ExampleCommandJob(CommandLineJob):

    def _setup(self):
        print 'Do not need to add args for this command!'
        self.arguments = ''

    def _get_command(self):
        return 'date'


# A template for a placeholder final job to add to the end of
# workflows.
FINAL_JOB = CommandJobTemplate('final', 'echo success')



# Copyright 2015, Pinterest, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Validation tests for blessed version."""
# We don't use 'testing' module since the project will likely move to its own
# repository.
import unittest

from pinball.master.blessed_version import BlessedVersion


__author__ = 'Pawel Garbacki'
__copyright__ = 'Copyright 2015, Pinterest, Inc.'
__credits__ = [__author__]
__license__ = 'Apache'
__version__ = '2.0'


class BlessedVersionTestCase(unittest.TestCase):
    def test_advance_version(self):
        blessed_version = BlessedVersion('some_name', 'some_owner')
        self.assertEqual('some_name', blessed_version.name)
        self.assertEqual('some_owner', blessed_version.owner)
        versions = []
        for _ in range(1, 1000):
            versions.append(blessed_version.advance_version())
        sorted_versions = list(versions)
        sorted_versions.sort()
        self.assertEqual(sorted_versions, versions)
